{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "# PyNEST Microcircuit: Network Class\n\nMain file of the microcircuit defining the ``Network`` class with functions to\nbuild and simulate the network.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import os\nimport numpy as np\nimport nest\nimport helpers\nimport warnings\n\n\nclass Network:\n    \"\"\" Provides functions to setup NEST, to create and connect all nodes of\n    the network, to simulate, and to evaluate the resulting spike data.\n\n    Instantiating a Network object derives dependent parameters and already\n    initializes the NEST kernel.\n\n    Parameters\n    ---------\n    sim_dict\n        Dictionary containing all parameters specific to the simulation\n        (see: ``sim_params.py``).\n    net_dict\n         Dictionary containing all parameters specific to the neuron and\n         network models (see: ``network_params.py``).\n    stim_dict\n        Optional dictionary containing all parameter specific to the stimulus\n        (see: ``stimulus_params.py``)\n\n    \"\"\"\n\n    def __init__(self, sim_dict, net_dict, stim_dict=None):\n        self.sim_dict = sim_dict\n        self.net_dict = net_dict\n        self.stim_dict = stim_dict\n\n        # data directory\n        self.data_path = sim_dict['data_path']\n        if nest.Rank() == 0:\n            if os.path.isdir(self.data_path):\n                message = '  Directory already existed.'\n                if self.sim_dict['overwrite_files']:\n                    message += ' Old data will be overwritten.'\n            else:\n                os.mkdir(self.data_path)\n                message = '  Directory has been created.'\n            print('Data will be written to: {}\\n{}\\n'.format(self.data_path,\n                                                             message))\n\n        # derive parameters based on input dictionaries\n        self.__derive_parameters()\n\n        # initialize the NEST kernel\n        self.__setup_nest()\n\n    def create(self):\n        \"\"\" Creates all network nodes.\n\n        Neuronal populations and recording and stimulation devices are created.\n\n        \"\"\"\n        self.__create_neuronal_populations()\n        if len(self.sim_dict['rec_dev']) > 0:\n            self.__create_recording_devices()\n        if self.net_dict['poisson_input']:\n            self.__create_poisson_bg_input()\n        if self.stim_dict['thalamic_input']:\n            self.__create_thalamic_stim_input()\n        if self.stim_dict['dc_input']:\n            self.__create_dc_stim_input()\n\n    def connect(self):\n        \"\"\" Connects the network.\n\n        Recurrent connections among neurons of the neuronal populations are\n        established, and recording and stimulation devices are connected.\n\n        The ``self.__connect_*()`` functions use ``nest.Connect()`` calls which\n        set up the postsynaptic connectivity.\n        Since the introduction of the 5g kernel in NEST 2.16.0 the full\n        connection infrastructure including presynaptic connectivity is set up\n        afterwards in the preparation phase of the simulation.\n        The preparation phase is usually induced by the first\n        ``nest.Simulate()`` call.\n        For including this phase in measurements of the connection time,\n        we induce it here explicitly by calling ``nest.Prepare()``.\n\n        \"\"\"\n        self.__connect_neuronal_populations()\n\n        if len(self.sim_dict['rec_dev']) > 0:\n            self.__connect_recording_devices()\n        if self.net_dict['poisson_input']:\n            self.__connect_poisson_bg_input()\n        if self.stim_dict['thalamic_input']:\n            self.__connect_thalamic_stim_input()\n        if self.stim_dict['dc_input']:\n            self.__connect_dc_stim_input()\n\n        nest.Prepare()\n        nest.Cleanup()\n\n    def simulate(self, t_sim):\n        \"\"\" Simulates the microcircuit.\n\n        Parameters\n        ----------\n        t_sim\n            Simulation time (in ms).\n\n        \"\"\"\n        if nest.Rank() == 0:\n            print('Simulating {} ms.'.format(t_sim))\n\n        nest.Simulate(t_sim)\n\n    def evaluate(self, raster_plot_interval, firing_rates_interval):\n        \"\"\" Displays simulation results.\n\n        Creates a spike raster plot.\n        Calculates the firing rate of each population and displays them as a\n        box plot.\n\n        Parameters\n        ----------\n        raster_plot_interval\n            Times (in ms) to start and stop loading spike times for raster plot\n            (included).\n        firing_rates_interval\n            Times (in ms) to start and stop lading spike times for computing\n            firing rates (included).\n\n        Returns\n        -------\n            None\n\n        \"\"\"\n        if nest.Rank() == 0:\n            print('Interval to plot spikes: {} ms'.format(raster_plot_interval))\n            helpers.plot_raster(\n                self.data_path,\n                'spike_recorder',\n                raster_plot_interval[0],\n                raster_plot_interval[1],\n                self.net_dict['N_scaling'])\n\n            print('Interval to compute firing rates: {} ms'.format(\n                firing_rates_interval))\n            helpers.firing_rates(\n                self.data_path, 'spike_recorder',\n                firing_rates_interval[0], firing_rates_interval[1])\n            helpers.boxplot(self.data_path, self.net_dict['populations'])\n\n    def __derive_parameters(self):\n        \"\"\"\n        Derives and adjusts parameters and stores them as class attributes.\n        \"\"\"\n        self.num_pops = len(self.net_dict['populations'])\n\n        # total number of synapses between neuronal populations before scaling\n        full_num_synapses = helpers.num_synapses_from_conn_probs(\n            self.net_dict['conn_probs'],\n            self.net_dict['full_num_neurons'],\n            self.net_dict['full_num_neurons'])\n\n        # scaled numbers of neurons and synapses\n        self.num_neurons = np.round((self.net_dict['full_num_neurons'] *\n                                     self.net_dict['N_scaling'])).astype(int)\n        self.num_synapses = np.round((full_num_synapses *\n                                      self.net_dict['N_scaling'] *\n                                      self.net_dict['K_scaling'])).astype(int)\n        self.ext_indegrees = np.round((self.net_dict['K_ext'] *\n                                       self.net_dict['K_scaling'])).astype(int)\n\n        # conversion from PSPs to PSCs\n        PSC_over_PSP = helpers.postsynaptic_potential_to_current(\n            self.net_dict['neuron_params']['C_m'],\n            self.net_dict['neuron_params']['tau_m'],\n            self.net_dict['neuron_params']['tau_syn'])\n        PSC_matrix_mean = self.net_dict['PSP_matrix_mean'] * PSC_over_PSP\n        PSC_ext = self.net_dict['PSP_exc_mean'] * PSC_over_PSP\n\n        # DC input compensates for potentially missing Poisson input\n        if self.net_dict['poisson_input']:\n            DC_amp = np.zeros(self.num_pops)\n        else:\n            if nest.Rank() == 0:\n                warnings.warn('DC input created to compensate missing Poisson input.\\n')\n            DC_amp = helpers.dc_input_compensating_poisson(\n                self.net_dict['bg_rate'], self.net_dict['K_ext'],\n                self.net_dict['neuron_params']['tau_syn'],\n                PSC_ext)\n\n        # adjust weights and DC amplitude if the indegree is scaled\n        if self.net_dict['K_scaling'] != 1:\n            PSC_matrix_mean, PSC_ext, DC_amp = \\\n                helpers.adjust_weights_and_input_to_synapse_scaling(\n                    self.net_dict['full_num_neurons'],\n                    full_num_synapses, self.net_dict['K_scaling'],\n                    PSC_matrix_mean, PSC_ext,\n                    self.net_dict['neuron_params']['tau_syn'],\n                    self.net_dict['full_mean_rates'],\n                    DC_amp,\n                    self.net_dict['poisson_input'],\n                    self.net_dict['bg_rate'], self.net_dict['K_ext'])\n\n        # store final parameters as class attributes\n        self.weight_matrix_mean = PSC_matrix_mean\n        self.weight_ext = PSC_ext\n        self.DC_amp = DC_amp\n\n        # thalamic input\n        if self.stim_dict['thalamic_input']:\n            num_th_synapses = helpers.num_synapses_from_conn_probs(\n                self.stim_dict['conn_probs_th'],\n                self.stim_dict['num_th_neurons'],\n                self.net_dict['full_num_neurons'])[0]\n            self.weight_th = self.stim_dict['PSP_th'] * PSC_over_PSP\n            if self.net_dict['K_scaling'] != 1:\n                num_th_synapses *= self.net_dict['K_scaling']\n                self.weight_th /= np.sqrt(self.net_dict['K_scaling'])\n            self.num_th_synapses = np.round(num_th_synapses).astype(int)\n\n        if nest.Rank() == 0:\n            message = ''\n            if self.net_dict['N_scaling'] != 1:\n                message += \\\n                    'Neuron numbers are scaled by a factor of {:.3f}.\\n'.format(\n                        self.net_dict['N_scaling'])\n            if self.net_dict['K_scaling'] != 1:\n                message += \\\n                    'Indegrees are scaled by a factor of {:.3f}.'.format(\n                        self.net_dict['K_scaling'])\n                message += '\\n  Weights and DC input are adjusted to compensate.\\n'\n            print(message)\n\n    def __setup_nest(self):\n        \"\"\" Initializes the NEST kernel.\n\n        Reset the NEST kernel and pass parameters to it.\n        \"\"\"\n        nest.ResetKernel()\n\n        nest.local_num_threads = self.sim_dict['local_num_threads']\n        nest.resolution = self.sim_dict['sim_resolution']\n        nest.rng_seed = self.sim_dict['rng_seed']\n        nest.overwrite_files = self.sim_dict['overwrite_files']\n        nest.print_time = self.sim_dict['print_time']\n        \n        if nest.Rank() == 0:\n            print('RNG seed: {}'.format(\n                nest.rng_seed))\n            print('Total number of virtual processes: {}'.format(\n                nest.total_num_virtual_procs))\n\n    def __create_neuronal_populations(self):\n        \"\"\" Creates the neuronal populations.\n\n        The neuronal populations are created and the parameters are assigned\n        to them. The initial membrane potential of the neurons is drawn from\n        normal distributions dependent on the parameter ``V0_type``.\n\n        The first and last neuron id of each population is written to file.\n        \"\"\"\n        if nest.Rank() == 0:\n            print('Creating neuronal populations.')\n\n        self.pops = []\n        for i in np.arange(self.num_pops):\n            population = nest.Create(self.net_dict['neuron_model'],\n                                     self.num_neurons[i])\n\n            population.set(\n                tau_syn_ex=self.net_dict['neuron_params']['tau_syn'],\n                tau_syn_in=self.net_dict['neuron_params']['tau_syn'],\n                E_L=self.net_dict['neuron_params']['E_L'],\n                V_th=self.net_dict['neuron_params']['V_th'],\n                V_reset=self.net_dict['neuron_params']['V_reset'],\n                t_ref=self.net_dict['neuron_params']['t_ref'],\n                I_e=self.DC_amp[i])\n\n            if self.net_dict['V0_type'] == 'optimized':\n                population.set(V_m=nest.random.normal(\n                    self.net_dict['neuron_params']['V0_mean']['optimized'][i],\n                    self.net_dict['neuron_params']['V0_std']['optimized'][i]))\n            elif self.net_dict['V0_type'] == 'original':\n                population.set(V_m=nest.random.normal(\n                    self.net_dict['neuron_params']['V0_mean']['original'],\n                    self.net_dict['neuron_params']['V0_std']['original']))\n            else:\n                raise ValueError(\n                    'V0_type is incorrect. ' +\n                    'Valid options are \"optimized\" and \"original\".')\n\n            self.pops.append(population)\n\n        # write node ids to file\n        if nest.Rank() == 0:\n            fn = os.path.join(self.data_path, 'population_nodeids.dat')\n            with open(fn, 'w+') as f:\n                for pop in self.pops:\n                    f.write('{} {}\\n'.format(pop[0].global_id,\n                                             pop[-1].global_id))\n\n    def __create_recording_devices(self):\n        \"\"\" Creates one recording device of each kind per population.\n\n        Only devices which are given in ``sim_dict['rec_dev']`` are created.\n\n        \"\"\"\n        if nest.Rank() == 0:\n            print('Creating recording devices.')\n\n        if 'spike_recorder' in self.sim_dict['rec_dev']:\n            if nest.Rank() == 0:\n                print('  Creating spike recorders.')\n            sd_dict = {'record_to': 'ascii',\n                       'label': os.path.join(self.data_path, 'spike_recorder')}\n            self.spike_recorders = nest.Create('spike_recorder',\n                                               n=self.num_pops,\n                                               params=sd_dict)\n\n        if 'voltmeter' in self.sim_dict['rec_dev']:\n            if nest.Rank() == 0:\n                print('  Creating voltmeters.')\n            vm_dict = {'interval': self.sim_dict['rec_V_int'],\n                       'record_to': 'ascii',\n                       'record_from': ['V_m'],\n                       'label': os.path.join(self.data_path, 'voltmeter')}\n            self.voltmeters = nest.Create('voltmeter',\n                                          n=self.num_pops,\n                                          params=vm_dict)\n\n    def __create_poisson_bg_input(self):\n        \"\"\" Creates the Poisson generators for ongoing background input if\n        specified in ``network_params.py``.\n\n        If ``poisson_input`` is ``False``, DC input is applied for compensation\n        in ``create_neuronal_populations()``.\n\n        \"\"\"\n        if nest.Rank() == 0:\n            print('Creating Poisson generators for background input.')\n\n        self.poisson_bg_input = nest.Create('poisson_generator',\n                                            n=self.num_pops)\n        self.poisson_bg_input.rate = \\\n            self.net_dict['bg_rate'] * self.ext_indegrees\n\n    def __create_thalamic_stim_input(self):\n        \"\"\" Creates the thalamic neuronal population if specified in\n        ``stim_dict``.\n\n        Each neuron of the thalamic population is supposed to transmit the same\n        Poisson spike train to all of its targets in the cortical neuronal population,\n        and spike trains elicited by different thalamic neurons should be statistically\n        independent.\n        In NEST, this is achieved with a single Poisson generator connected to all\n        thalamic neurons which are of type ``parrot_neuron``;\n        Poisson generators send independent spike trains to each of their targets and\n        parrot neurons just repeat incoming spikes.        \n        \n        Note that the number of thalamic neurons is not scaled with\n        ``N_scaling``.\n\n        \"\"\"\n        if nest.Rank() == 0:\n            print('Creating thalamic input for external stimulation.')\n\n        self.thalamic_population = nest.Create(\n            'parrot_neuron', n=self.stim_dict['num_th_neurons'])\n\n        self.poisson_th = nest.Create('poisson_generator')\n        self.poisson_th.set(\n            rate=self.stim_dict['th_rate'],\n            start=self.stim_dict['th_start'],\n            stop=(self.stim_dict['th_start'] + self.stim_dict['th_duration']))\n\n    def __create_dc_stim_input(self):\n        \"\"\" Creates DC generators for external stimulation if specified\n        in ``stim_dict``.\n\n        The final amplitude is the ``stim_dict['dc_amp'] * net_dict['K_ext']``.\n\n        \"\"\"\n        dc_amp_stim = self.stim_dict['dc_amp'] * self.net_dict['K_ext']\n\n        if nest.Rank() == 0:\n            print('Creating DC generators for external stimulation.')\n\n        dc_dict = {'amplitude': dc_amp_stim,\n                   'start': self.stim_dict['dc_start'],\n                   'stop': self.stim_dict['dc_start'] + self.stim_dict['dc_dur']}\n        self.dc_stim_input = nest.Create('dc_generator', n=self.num_pops, params=dc_dict)\n\n    def __connect_neuronal_populations(self):\n        \"\"\" Creates the recurrent connections between neuronal populations. \"\"\"\n        if nest.Rank() == 0:\n            print('Connecting neuronal populations recurrently.')\n\n        for i, target_pop in enumerate(self.pops):\n            for j, source_pop in enumerate(self.pops):\n                if self.num_synapses[i][j] >= 0.:\n                    conn_dict_rec = {\n                        'rule': 'fixed_total_number',\n                        'N': self.num_synapses[i][j]}\n\n                    if self.weight_matrix_mean[i][j] < 0:\n                        w_min = np.NINF\n                        w_max = 0.0\n                    else:\n                        w_min = 0.0\n                        w_max = np.Inf\n\n                    syn_dict = {\n                        'synapse_model': 'static_synapse',\n                        'weight': nest.math.redraw(\n                            nest.random.normal(\n                                mean=self.weight_matrix_mean[i][j],\n                                std=abs(self.weight_matrix_mean[i][j] *\n                                        self.net_dict['weight_rel_std'])),\n                            min=w_min,\n                            max=w_max),\n                        'delay': nest.math.redraw(\n                            nest.random.normal(\n                                mean=self.net_dict['delay_matrix_mean'][i][j],\n                                std=(self.net_dict['delay_matrix_mean'][i][j] *\n                                     self.net_dict['delay_rel_std'])),\n                            min=nest.resolution,\n                            max=np.Inf)}\n\n                    nest.Connect(\n                        source_pop, target_pop,\n                        conn_spec=conn_dict_rec,\n                        syn_spec=syn_dict)\n\n    def __connect_recording_devices(self):\n        \"\"\" Connects the recording devices to the microcircuit.\"\"\"\n        if nest.Rank == 0:\n            print('Connecting recording devices.')\n\n        for i, target_pop in enumerate(self.pops):\n            if 'spike_recorder' in self.sim_dict['rec_dev']:\n                nest.Connect(target_pop, self.spike_recorders[i])\n            if 'voltmeter' in self.sim_dict['rec_dev']:\n                nest.Connect(self.voltmeters[i], target_pop)\n\n    def __connect_poisson_bg_input(self):\n        \"\"\" Connects the Poisson generators to the microcircuit.\"\"\"\n        if nest.Rank() == 0:\n            print('Connecting Poisson generators for background input.')\n\n        for i, target_pop in enumerate(self.pops):\n            conn_dict_poisson = {'rule': 'all_to_all'}\n\n            syn_dict_poisson = {\n                'synapse_model': 'static_synapse',\n                'weight': self.weight_ext,\n                'delay': self.net_dict['delay_poisson']}\n\n            nest.Connect(\n                self.poisson_bg_input[i], target_pop,\n                conn_spec=conn_dict_poisson,\n                syn_spec=syn_dict_poisson)\n\n    def __connect_thalamic_stim_input(self):\n        \"\"\" Connects the thalamic input to the neuronal populations.\"\"\"\n        if nest.Rank() == 0:\n            print('Connecting thalamic input.')\n\n        # connect Poisson input to thalamic population\n        nest.Connect(self.poisson_th, self.thalamic_population)\n\n        # connect thalamic population to neuronal populations\n        for i, target_pop in enumerate(self.pops):\n            conn_dict_th = {\n                'rule': 'fixed_total_number',\n                'N': self.num_th_synapses[i]}\n\n            syn_dict_th = {\n                'weight': nest.math.redraw(\n                    nest.random.normal(\n                        mean=self.weight_th,\n                        std=self.weight_th * self.net_dict['weight_rel_std']),\n                    min=0.0,\n                    max=np.Inf),\n                'delay': nest.math.redraw(\n                    nest.random.normal(\n                        mean=self.stim_dict['delay_th_mean'],\n                        std=(self.stim_dict['delay_th_mean'] *\n                             self.stim_dict['delay_th_rel_std'])),\n                    min=nest.resolution,\n                    max=np.Inf)}\n\n            nest.Connect(\n                self.thalamic_population, target_pop,\n                conn_spec=conn_dict_th, syn_spec=syn_dict_th)\n\n    def __connect_dc_stim_input(self):\n        \"\"\" Connects the DC generators to the neuronal populations. \"\"\"\n\n        if nest.Rank() == 0:\n            print('Connecting DC generators.')\n\n        for i, target_pop in enumerate(self.pops):\n            nest.Connect(self.dc_stim_input[i], target_pop)"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.8.6"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}